#!/usr/bin/env ruby_devops
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# Author::     Ravi Shankar (ravi_sabapathy@dell.com)
# Copyright::  Copyright (c) 2017, Dell Inc. All rights reserved.
# License::    [Apache License] (http://www.apache.org/licenses/LICENSE-2.0)
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file contains OS10 shell implementation.

require 'socket'
require 'fileutils'
require 'logger'
require 'xml/libxml'
require 'timeout'
require 'open3'

# Temporary files to process standard output and error
$connection_handle = nil
PROCESS_OUTPUT = 'Read output'.freeze
XML_CMD = 'display-xml'.freeze
ERROR_STRING = 'Syntax error '.freeze
FILE_PATH = '/tmp/'.freeze
SEND_OUTPUT = 'Read output\n'.freeze

$os10_shell_log = Logger.new(STDOUT)

# format the log output
$os10_shell_log.formatter = proc do |severity, datetime, progname, msg|
  date_format = datetime.strftime('%Y-%m-%d %H:%M:%S')
  if %w[INFO WARN].include?(severity)
    "[#{date_format}] #{severity}  (#{progname}): #{msg}\n"
  else
    "[#{date_format}] #{severity} (#{progname}): #{msg}\n"
  end
end

# Clear the temporary files
def cleanup_tmp_file
  out_path = FILE_PATH + 'out.txt'
  err_path = FILE_PATH + 'error.txt'
  std_output_file = FILE_PATH + 'stdout.txt'
  std_error_file = FILE_PATH + 'stderr.txt'

  File.truncate(out_path, 0) if File.exist?(out_path)
  File.truncate(err_path, 0) if File.exist?(err_path)
  File.truncate(std_output_file, 0) if File.exist?(std_output_file)
  File.truncate(std_error_file, 0) if File.exist?(std_error_file)
end

# The XmlToHash class has utility api to validate the given
# XML format and convert the XML to ruby hash
class XmlToHash
  class << self
    # This API validates xml node and convert to ruby hash
    def from_libxml(xml, strict = true)
      begin
        XML.default_load_external_dtd = false
        XML.default_pedantic_parser = strict
        result = XML::Parser.string(xml).parse
        return { result.root.name.to_s => xml_node_to_hash(result.root) }
      rescue StandardError => e
        #cleanup_tmp_file
        # raise your custom exception here
        raise "xml to ruby hash conversion error: #{e}"
      end
    end

    # Api to convert xml node to ruby hash.
    # this api will recusively call the sub nodes of xml.
    def xml_node_to_hash(node)
      # If we are at the root of the document, start the hash
      if node.element?
        if node.children?
          result_hash = {}

          node.each_child do |child|
            result = xml_node_to_hash(child)

            if child.name == 'text'
              return result if !child.next? && !child.prev?
            elsif result_hash[child.name.to_sym]
              if result_hash[child.name.to_sym].is_a?(Object::Array)
                result_hash[child.name.to_sym] << result
              else
                result_hash[child.name.to_sym] =
                  [result_hash[child.name.to_sym]] << result
              end
            else
              result_hash[child.name.to_sym] = result
            end
          end
          return result_hash
        else
          return nil
        end
      else
        return node.content.to_s
      end
    end
  end
end

def get_sub_child_pid(pid)
  if pid > 1
    out, err, status = Open3.capture3('pgrep -P ' + pid.to_s + '|xargs')
    return out.to_i
  end
  0
end

def is_user_present(username)
  out, err, status = Open3.capture3('id -Gn ' + username)
  raise err unless err.empty?
  unless out.empty?
    group_string = out.split(' ')
    return true if group_string[0] == 'sysadmin'
    raise 'user ' + username + ' should be created in the group sysadmin'
  end
end  

# Check if a process is alive or not by executing the
# netstat command. It gets the current process from netstat
# command and checks if it has a child process. If the child
# and parent process are present then the OS10 shell process is alive
#
# params:
# +hostname+::     ip address in string format
# +port+::         port to bind for os10shell
#
# Return:
# +isalive+::    process is running or not
#
# Raises:
#    None

def is_process_alive(hostname, port)
  netstat_cmd = "netstat -nlp | grep #{hostname}:" + \
                port.to_s + ' | grep LISTEN'
  out, err, status = Open3.capture3(netstat_cmd)
  unless out.empty?
    parts = out.split(/\s+/)
    pid_str = parts[6]
    pid = pid_str.split('/')
    paren_pid = pid[0].to_i
    return { parent_pid: 0, child_pid: 0 } unless paren_pid
    stdout, stderr, stat = Open3.capture3('pgrep -P ' + pid[0] + '|xargs')
    child_pid = stdout.to_i
    sub_child_pid = get_sub_child_pid(child_pid)
    if sub_child_pid > 1
      last_child_pid = get_sub_child_pid(sub_child_pid)
      return { parent_pid: paren_pid, child_pid: child_pid } if last_child_pid > 1
    end
    if sub_child_pid.zero? || last_child_pid.zero?
      Process.kill('KILL', child_pid)
      Process.kill('KILL', paren_pid)
    end
  end
  { parent_pid: 0, child_pid: 0 }
end

# Execute the shell script in a seperate process and detach from it
#
# params:
# +command+::    shell command to execute
#
# Return:
# +pid+::        pid of the shell command
#
# Raises:
#    None

def start(command)
  pid = spawn(command)
  Process.detach(pid)
  sleep 2
end

# Start the os10 shell only if it is not running and
# creates client socket.
#
# params:
#
# +username+::     username credentials
# +hostname+::     ip address in string format
# +port+::         port to bind for os10shell
#
# Return:
#    None
#
# Raises:
#    None

def start_os10_shell(username = 'os10devops',
                     hostname = '127.0.0.1',
                     port = 20_000)
  username = 'os10devops' if username != 'os10devops'
  begin
    is_user_present('os10devops')
    pid_hash = is_process_alive(hostname, port)
    if pid_hash[:parent_pid].zero? && pid_hash[:child_pid].zero?
      # spawn the os10shell process only if it is not spawned already
      shell_cmd = '/opt/dell/os10/bin/devops/dellos10_shell_process.rb ' + \
                  username + ' ' + ' ' + hostname + ' ' + port.to_s
      start(shell_cmd)
      $connection_handle = create_conn_handle(hostname, port, 20, 1)
    else
      if $connection_handle.nil?
        $connection_handle = create_conn_handle(hostname, port, 20, 1)
      end
    end
    raise 'connection handle is null' if $connection_handle == nil
  rescue Errno::ESRCH
    return
  end
end

# The api exits the OS10 shell and cleans up the temp files
#
# params:
#    None
#
# Return:
# +result+::    hash map of command output and error
#
# Raises:
#    None

def end_os10_shell
  exit_shell_cmd = Array.new(["exit"])
  result = execute_exec_command(exit_shell_cmd)
  unless $connection_handle.nil?
    $connection_handle.close
    $connection_handle = nil
  end
  result
end

# The Utility API closes the TCP socket and
# returns false during error
#
# params:
#
# +hostname+::     ip address in string format
# +port+::         port to bind for os10shell
#
# Return:
# +status+::       true if port is open
#                  false if error is encountered
# Raises:
# +Errno+::        ECONNREFUSED, EHOSTUNREACH

def close_open_port(host, port)
  begin
    TCPSocket.new(host, port).close
    return true
  rescue Errno::ECONNREFUSED, Errno::EHOSTUNREACH
    return false
  end
  false
end

# Creates the TCP client socket with the TCP server
# running in OS10 shell.
#
# params:
#
# +hostname+::     ip address in string format
# +port+::         port to bind for os10shell
# +timeout+::      Number of times to retry
# +sleep_period+:: timeout value
#
# Return:
# +status+::      nil on Timeout error
#
# Raises:
# +Error+::              Timeout

def create_conn_handle(hostname, port, timeout, sleep_period)
  begin
    Timeout::timeout(timeout) do
      begin
        return TCPSocket.new hostname, port
      rescue Errno::ECONNREFUSED, Errno::EHOSTUNREACH
        sleep(sleep_period)
        retry
      end
    end
  rescue Timeout::Error
    return nil
  end
end

# Read content from the given file
#
# params:
# +file_name+::  File to be read
#
# Return:
# +data+::       read buffer
#
# Raises:
#    None

def read_file(file_name)
  file = File.open(file_name, 'rb')
  data = file.read
  file.close
  data
end

# Few errors are printed in the stdout file. Parse
# such errors and raise exception. This API is currently
# used by execute_exec_command
#
# params:
# +fstdout_str+::   command output from shell process
#
# Return:
#     None
#
# Raises:
#    StandardError  is raised if error is found

def find_error(stdout_str)
  return if stdout_str.nil? || stdout_str.empty?
  arr = stdout_str.split("\n")
  size = arr.size
  array_size = size - 1
  (0..array_size).each do |index|
    # Check if error is present in the file
    next if arr[index].nil?
    if arr[index].include? "% Error:"
      if index != 0
        command = arr[index - 1].split('#')
        unless command[1].nil?
          $os10_shell_log.error 'Error in command - ' + command[1]
        end
      end
      cleanup_tmp_file
      raise arr[index]
    end
  end
end

# Remove the junk characters from the os10 shell.
# The output obtained from the os10 shell contains NULL
# characters and os10 prompt strings. These characters
# have to be removed for further processing.
#
# params:
# +file_path+::  File to be read
# +command+::    Cli command given by user
#
# Return:
# +arr+::i       read buffer without NULL and OS10 prompt
#
# Raises:
#    None

def remove_junk_file_chars(file_path, command)
  arr = IO.readlines(file_path)
  length = arr.size
  return '' if length.zero?
  end_loop = false
  arr[length - 1].replace('')
  if command.include? XML_CMD
    # Remove the os10 prompt shell from the output
    start_xml_tag = false
    end_xml_tag = false
    (0..length).each do |index|
      next if arr[index].nil?
      unless start_xml_tag
        start_xml_tag = true if arr[index].include? '<rpc-reply>'
        arr[index].replace('') unless start_xml_tag
      end
      if start_xml_tag && !end_xml_tag
        end_xml_tag = true if arr[index].include? '</rpc-reply>'
        next
      end
      arr[index].replace('') if end_xml_tag
    end
    return '' if !end_xml_tag || !start_xml_tag
  else
    (0..1).each do |index|
      # Remove the NULL chacters from the output
      end_loop = true if index.zero? && arr[index][0] != "\0"
      arr[index].replace('')
      break if end_loop == true
    end
  end
  arr.join('')
end

# Utility API to receive and process output
#
# params:
# +command+::        show command to be executed
# +show_cmd+::       true for show command and false for other commands
#
# Return:
# +stdout+::         console output from the shell
# +stderr+::         console error output from the shell
#
# Raises:
#   None

def process_output(command = '', show_cmd = false)

  std_output_file = FILE_PATH + 'stdout.txt'
  std_error_file = FILE_PATH + 'stderr.txt'

  # Process the stdout and stderr files after the server sends
  # a message
  msg = $connection_handle.gets
  if !msg.nil? and msg.include? PROCESS_OUTPUT
    if File.exist?(std_error_file)
      stderr = read_file(std_error_file)
      # replace the null character in command error
      stderr.gsub!(/\0/, '')
      if !stderr.empty? and stderr.include? ERROR_STRING
        cleanup_tmp_file
        raise stderr
      end
    end
    if File.exist?(std_output_file)
      if show_cmd == true
        if command.nil?
          stdout = ''
        else
          temp = remove_junk_file_chars(std_output_file, command)
          return { stdout: '', stderr: stderr } if temp.empty?
          if command.include? XML_CMD
             stdout = XmlToHash.from_libxml(temp)
          else
            stdout = temp
          end
        end
      else
        stdout = read_file(std_output_file)
        # replace the null character in command output
        stdout.gsub!(/\0/, '')
      end
    end
  elsif msg.nil?
    # OS10 Shell server socket is closed
    $connection_handle.close
    $connection_handle = nil
    raise 'Server socket closed'
  end
  { stdout: stdout, stderr: stderr }
end

# Execute interactive/normal commands in the exec mode given by the user
#
# params:
# +command_array+::  list of commands to be executed
#
# Return:
# +result+::         hash map containing console output and error
#
# Raises:
# +Error+::          EPIPE

def execute_exec_command(command_array)
  return { stdout: '', stderr: 'Command array cannot be empty' } \
    if command_array.empty?

  command_array.each do |cmd|
    return { stdout: '', stderr: 'Command in array cannot be empty' } \
      if cmd.empty?
  end
  begin
    $connection_handle.write "Interactive Command\n"
    command_array.each do |cmd|
      $connection_handle.write cmd + "\r\n"
    end
    $connection_handle.write "\r\n"
    $connection_handle.flush
    result = process_output
  rescue Errno::EPIPE => e
    $connection_handle.close
    $connection_handle = nil
    cleanup_tmp_file
    raise e
  rescue Errno::ENOENT => e
    $connection_handle.close
    $connection_handle = nil
    result = { stdout: '', stderr: '' }
  end
  result
end

# Execute config the command given by the user
#
# params:
# +command_array+::  list of commands to be executed
# +timeout+::        wait for the shell to print the output
#
# Return:
# +result+::         hash map containing console output and error
#
# Raises:
# +Error+::          EPIPE

def execute_config_command(command_array, timeout = 12)
  return { stdout: '', stderr: 'Timeout should be an Integer' } \
    if !timeout.is_a? Integer
  return { stdout: '', stderr: 'Timeout should be >= 1' } if timeout < 1
  return { stdout: '', stderr: 'Command array cannot be empty' } \
    if command_array.empty?

  command_array.each do |cmd|
    return { stdout: '', stderr: 'Command in array cannot be empty' } \
      if cmd.empty?
  end
  result = { stdout: '', stderr: '' }
  begin
    $connection_handle.write "Timeout " + timeout.to_s + "\n"
    $connection_handle.write "configure terminal\n"
    command_array.each do |cmd|
      $connection_handle.write cmd + "\n"
    end
    $connection_handle.write "end\n\r\n"
    $connection_handle.flush
    result = process_output
    find_error(result[:stdout])
  rescue Errno::EPIPE => e
    $connection_handle.close
    $connection_handle = nil
    cleanup_tmp_file
  rescue Errno::ENOENT => e
    $connection_handle.close
    $connection_handle = nil
  end
  result
end

# Execute show command given by the user
#
# params:
# +command+::        show command to be executed
# +timeout+::        wait for the shell to print the output
#
# Return:
# +result+::         hash map containing console output and error
#
# Raises:
# +Error+::          EPIPE

def execute_show_command(command, timeout = 15)
  return { stdout: '', stderr: 'Timeout should be an Integer' } \
    unless timeout.is_a? Integer
  return { stdout: '', stderr: 'Timeout should be >= 1' } if timeout < 1
  return { stdout: '', stderr: 'Command cannot be empty' } if command.empty?

  result = { stdout: '', stderr: '' }
  begin
    $connection_handle.write "Timeout " + timeout.to_s + "\n"
    $connection_handle.write command + "\n\r\n"
    $connection_handle.flush
    result = process_output(command, true)
  rescue Errno::EPIPE => e
    $os10_shell_log.error e.message
    $connection_handle.close
    $connection_handle = nil
    cleanup_tmp_file
  rescue Errno::ENOENT => e
    $connection_handle.close
    $connection_handle = nil
  end
  result
end
