#!/usr/bin/env ruby_devops
require 'socket'
require 'fileutils'
require 'logger'
require 'timeout'

FILE_PATH = '/tmp/'.freeze
SEND_OUTPUT = "Read output\n".freeze

$os10_shell_log = Logger.new(STDOUT)

# format the log output
$os10_shell_log.formatter = proc do |severity, datetime, progname, msg|
  date_format = datetime.strftime('%Y-%m-%d %H:%M:%S')
  if %w[INFO WARN].include?(severity)
    "[#{date_format}] #{severity}  (#{progname}): #{msg}\n"
  else
    "[#{date_format}] #{severity} (#{progname}): #{msg}\n"
  end
end

# Cleanup temporary file
def cleanup_resources(write_io = nil,
                      child_pid = 0)
  out_path = FILE_PATH + 'out.txt'
  err_path = FILE_PATH + 'error.txt'
  std_output_file = FILE_PATH + 'stdout.txt'
  std_error_file = FILE_PATH + 'stderr.txt'

  File.delete(out_path) if File.file?(out_path)
  File.delete(err_path) if File.file?(err_path)
  File.delete(std_output_file) if File.file?(std_output_file)
  File.delete(std_error_file) if File.file?(std_error_file)
  if !write_io.nil? && !child_pid.zero?
    write_io.write "exit\r\n"
    Process.wait child_pid
  end
end

# Copy the file contents to a file
#
# params:
# +file_name+::  File to be read
# +buffer   +::  data to write
#
# Return:
#
# Raises:
#    None

def copy_file(to_file, buffer)
  dst_file = File.open(to_file, 'wb')
  dst_file.write(buffer)
  dst_file.close
end

# Read content from the given file
#
# params:
# +file_name+::  File to be read
#
# Return:
# +data+::       read buffer
#
# Raises:
#    None

def read_file(file_name)
  return '' unless File.file?(file_name)
  file = File.open(file_name, 'rb')
  data = file.read
  file.close
  data
end

# Utility API to check if the command is complete or not
#
# params:
# +file_name+::       file name
#                     timeout
# Return:
#    hash of file output and prompt count
#
# Raises:
#    None

def is_command_complete(file_name)#, host_str)
  arr = IO.readlines(file_name)
  length = arr.size
  return { data: '', prompt_cnt: 0 } if length == 0
  index = 0
  index = 1 if arr[0][0] == "\0"
  if !arr[index].nil?
    return { data: '', prompt_cnt: 2 } if arr[index].include?'terminal length 0'
    host_str = arr[index].split('#')
    command_prompt = host_str[0] + '#' 
    length = arr.size
    count = 1
    len = length - 1
    return { data: '', prompt_cnt: count } if len <= 1
    if index != len && !arr[len].nil?
      count += 1 if arr[len].include? command_prompt
    end
  else
    return { data: '', prompt_cnt: 0 }
  end
  { data: arr.join(''), prompt_cnt: count }
end

# Wait for data in the file and then read it
#
# params:
# +file_name+::       file name
# +timeout_value+::   wait for command for the given
#                     timeout
# Return:
#    None
#
# Raises:
#    None

def read_output(file_name, show_command,
                timeout_value = 1)
  file_size = 0
  old_file_size = 0
  dampen_cnt = 0
  valid_size_cnt = 0
  total_wait_cnt = 0
  begin
    timeout = (timeout_value * 5)
    loop do
      if total_wait_cnt >= timeout# 300
        output = is_command_complete(file_name)
        return output[:data] if output[:prompt_cnt] >= 2
        return ''
      end
          
      output = is_command_complete(file_name)
      return output[:data] if output[:prompt_cnt] >= 2     
      sleep 0.2
      total_wait_cnt += 1
    end
  rescue Errno::ENOENT => e
    File.new(file_name, 'wb').close
  end
end

# Send the command to the shell and get the command
# output from the shell
#
# params:
# +descriptors+::     socket in a array
# +newsock+::     new client socket
# +write_io+:     write pipe from parent to child
#
# Return:
# +is_exit+::     return true if exit is received

def process_sock_data(descriptors,
                      newsock,
                      write_io)
  out_path = FILE_PATH + 'out.txt'

  err_path = FILE_PATH + 'error.txt'
  std_output_file = FILE_PATH + 'stdout.txt'
  std_error_file = FILE_PATH + 'stderr.txt'

  is_exit = false
  interactive_command = false
  timeout_value = 1
  show_command = false
  unless write_io.nil?
    if !File.file?(out_path)
      File.new(out_path, 'wb').close
    else
      File.truncate(out_path, 0)
    end
    if !File.file?(err_path)
      File.new(err_path, 'wb').close
    else
      File.truncate(err_path, 0)
    end
    File.new(std_output_file, 'wb').close unless File.file?(std_output_file)
    File.new(std_error_file, 'wb').close unless File.file?(std_error_file)
    while (os10_cmd = newsock.gets) != "\r\n"
      if !os10_cmd.nil?
        if os10_cmd.include? "Timeout "
          timeout_value = os10_cmd.scan(/\d+/).first.to_i
          next
        elsif os10_cmd.include? "Interactive Command"
          interactive_command = true
          next
        elsif os10_cmd.include? "exit"
          is_exit = true if interactive_command == true
        elsif os10_cmd.include? "show "
          show_command = true
        end
        write_io.write os10_cmd
        sleep 1 if interactive_command
      else
        newsock.close
        descriptors.delete(newsock)
        return is_exit
      end
    end
    write_io.write "\n" if is_exit == false
    # Copy the stdout and strerr in a seperate file which
    # will be read by the client code
    data = read_output(out_path, show_command, timeout_value)
    copy_file(std_output_file, data)
    read_error = read_file(err_path)
    if !read_error.empty?
      copy_file(std_error_file, read_error)
    else
      File.truncate(std_error_file, 0) if File.file?(std_error_file)
    end
    newsock.write SEND_OUTPUT
    is_exit
  end
end

# Api to accept new client TCP connection and process data
# from client
#
# params:
# +descriptors+::     socket in a array
# +server_socket+::   server socket
# +write_io+::        write pipe from parent to child
# +child_pid+::       process ID shell process
#
# Return:
#    None
#
# Raises
# +Exception+::       General socket exception

def accept_new_connection(descriptors,
                          server_socket,
                          write_io,
                          child_pid)
  begin
    newsock = server_socket.accept
  rescue StandardError => e
    $os10_shell_log.error 'Tcp client socket accpet error: #{e}'
    cleanup_resources(write_io, child_pid)
    raise
  end
  descriptors.push(newsock)
  if process_sock_data(descriptors, newsock, write_io)
    cleanup(descriptors, server_socket, child_pid)
  end
end

# Clean all the sockets before exiting the parent process
#
# params:
# +descriptors+::     socket in a array
# +server_socket+::   server socket
# +child_pid+::       exec process ID
#
# Return:
#      None
#
# Raises:
#    None

def cleanup(descriptors,
            server_socket,
            child_pid)
  descriptors.each do |sock|
    # close all the client and server sockets
    if sock == server_socket
      server_socket.shutdown(Socket::SHUT_RDWR)
    end
    sock.close
    descriptors.delete(sock)
  end
  cleanup_resources
  Process.wait child_pid
  exit
end

# Wait for data from the client socket and then process
#
# params:
# +descriptors+::     socket in a array
# +server_socket+::   server socket
# +write_io+::        write pipe from parent to child
# +child_pid+::       exec process Id
#
# Return:
#      None
#
# Raises:
#    None

def run(descriptors,
        server_socket,
        write_io,
        child_pid)
  loop do
    res = select(descriptors)
    unless res.nil?
      for sock in res[0]
        if sock == server_socket then
          accept_new_connection(descriptors,
                                server_socket,
                                write_io,
                                child_pid)
        else
          if sock.eof? then
            sock.close
            descriptors.delete(sock)
          else
            if process_sock_data(descriptors, sock, write_io)
              cleanup(descriptors, server_socket, child_pid)
            end
          end
        end
      end
    end
  end
end

# Creates the TCP server socket with the TCP server
# running in OS10 shell.
#
# params:
#
# +hostname+::     ip address in string format
# +port+::         port to bind for os10shell
# +timeout+::      Number of times to retry
# +sleep_period+:: timeout value
#
# Return:
# +status+::      nil on Timeout error
#
# Raises:
# +Error+::              Timeout

def create_server_socket(host_name, port, timeout, sleep_period)
  begin
    Timeout.timeout(timeout) do
      begin
        return TCPServer.new host_name, port.to_s
      rescue Errno::EADDRINUSE
        sleep(sleep_period)
        retry
      end
    end
  rescue Timeout::Error
    return nil
  end
end

# Creates the os10shell process with all the communication
# mechanism with the exec process. It stores the outputs
# in a file.
#
# params:
#
# +username+::     username of the shell
# +hostname+::     ip address in string format
# +port+::         port to bind for os10shell
#
# Return:
#  None
#
# Raises:
# +Errno+::              EPIPE

def spawn_os10shell(user_name, host_name, port_num)
  begin
    out_path = FILE_PATH + 'out.txt'
    err_path = FILE_PATH + 'error.txt'
    std_output_file = FILE_PATH + 'stdout.txt'
    std_error_file = FILE_PATH + 'stderr.txt'

    server_socket = create_server_socket(host_name, port_num, 5, 1)
    if server_socket.nil?
      raise 'Error in server socket creation'
    end

    # create a pipe
    read_io, write_io = IO.pipe

    # Create file required for shell process
    File.new(out_path, 'wb').close
    File.new(err_path, 'wb').close
    File.new(std_output_file, 'wb').close
    File.new(std_error_file, 'wb').close

    sleep 0.5

    child_pid = fork do
      # In the child process exec the su - username shell

      # close the server socket
      server_socket.close

      # close the write end of the pipe
      write_io.close
      $stdout.reopen(out_path, 'wb')
      $stderr.reopen(err_path, 'wb')

      # change our stdout to be the read end of the pipe
      STDIN.reopen(read_io)
      exec_str = 'su - ' + user_name
      exec exec_str
    end

    read_io.close

    # In case sudo - username asks for password enter it
    # write_io.write passwd
    # write_io.write '\n'
    # set terminal length to get the whole output from the shell
    write_io.write "terminal length 0\r\n"
    write_io.write "\n"
    read_output(out_path, false, 10)

    # write what we want to the pipe, it will be sent to childs stdout
    descriptors = []
    descriptors.push(server_socket)
    run(descriptors, server_socket, write_io, child_pid)
  rescue Errno::EPIPE => e
    cleanup_resources(write_io, child_pid)
    raise e
  rescue Errno::ENOENT => e
    cleanup_resources(write_io, child_pid)
    raise e
  rescue Errno::EROFS => e
    cleanup_resources(write_io, child_pid)
    raise e
  rescue StandardError => e
    cleanup_resources
    raise e
  end
end

if ARGV[0].nil? || ARGV[1].nil? || ARGV[2].nil?
  $os10_shell_log.error 'Execute the script with all mandatory parameters: ' + \
                        'username, hostname and port'
  exit
end

spawn_os10shell(ARGV[0], ARGV[1], ARGV[2])
